#include <go32.h>
#include <dpmi.h>
#include <pc.h>
#include <stdlib.h>
#include <dos.h>
#include "comport.h"

#include <iostream.h>

char XOFF = 0x13;
char XON = 0x11;

CComPort * pComPort = NULL;

int PortOffsets[] = 
{ 0, 1, 0, 0, 1, 3, 4, 5, 6 };

// -----------------------------
// Communication port interrupt handler 
// -----------------------------
void ComPortHandler()
{
  char c = inportb(pComPort->m_Regs8250.RBR);
#ifdef COM_PORT_XON_XOFF_ENABLED
  if (c == XOFF)
    pComPort->m_bFlowSend = 0;
  else if (c == XON)
    pComPort->m_bFlowSend = 1;
  else
  {
#endif
    pComPort->m_Buffer[pComPort->m_BufferHead] = c;
    pComPort->m_BufferHead = (pComPort->m_BufferHead + 1) % COM_PORT_BUFFER_SIZE;
#ifdef COM_PORT_XON_XOFF_ENABLED
    if (pComPort->m_bFlowRecv && pComPort->BufferLength() > COM_PORT_BUFFER_FLOW_OFF)
    {
      pComPort->WriteData(&XOFF);
      pComPort->m_bFlowRecv = 0;
    }
  }
#endif
  __asm__ __volatile__ ("cli");
  outportb(0x20, 0x20);
}

CComPort::CComPort(int nComPort = 1, int BasePort, int IRQ, int Interrupt)
{
  switch (nComPort)
  {
  case 1:
    {
      BasePort = 0x3f8;
      Interrupt = 0x0c;
      IRQ = 4;
    }
    break;
  case 2:
    {
      BasePort = 0x2f8;
      Interrupt = 0x0b;
      IRQ = 3;
    }
    break;
  case 3:
    {
      BasePort = 0x3e8;
      Interrupt = 0x0c;
      IRQ = 4;
    }
    break;
  case 4:
    {
      BasePort = 0x2e8;
      Interrupt = 0x0b;
      IRQ = 3;
    }
    break;
  default:
    {
    }
  }
  m_nComPort = nComPort;
  Create(BasePort, IRQ, Interrupt);
}

void CComPort::Create(int BasePort, int IRQ, int Interrupt)
{
  pComPort = this;
  m_Regs8250.DLL = BasePort + PortOffsets[0];
  m_Regs8250.DLH = BasePort + PortOffsets[1];
  m_Regs8250.THR = BasePort + PortOffsets[2];
  m_Regs8250.RBR = BasePort + PortOffsets[3];
  m_Regs8250.IER = BasePort + PortOffsets[4];
  m_Regs8250.LCR = BasePort + PortOffsets[5];
  m_Regs8250.MCR = BasePort + PortOffsets[6];
  m_Regs8250.LSR = BasePort + PortOffsets[7];
  m_Regs8250.MSR = BasePort + PortOffsets[8];
  m_nInterrupt = Interrupt;
  m_nIRQ = IRQ;
  m_nBasePort = BasePort;  

  // Reset serial port
  outportb(m_Regs8250.MCR, 0x00);
  outportb(m_Regs8250.LCR, 0x00);
  outportb(m_Regs8250.IER, 0x00);
  outportb(m_Regs8250.MSR, 0x00);
  outportb(m_Regs8250.LSR, 0x00);
  outportb(m_Regs8250.LCR, 0x80); 
  outportb(m_Regs8250.DLH, 0x00);
  outportb(m_Regs8250.DLL, 0x00);
  outportb(m_Regs8250.LCR, 0x00);
  
  // Set up the handler 
  m_BufferHead = m_BufferTail = 0;
  m_NewInfo.pm_offset = (int)ComPortHandler;
  m_NewInfo.pm_selector = _go32_my_cs();
  _go32_dpmi_get_protected_mode_interrupt_vector(m_nInterrupt, &m_OldInfo);
  _go32_dpmi_allocate_iret_wrapper(&m_NewInfo);
  _go32_dpmi_set_protected_mode_interrupt_vector(m_nInterrupt, &m_NewInfo);

  // Modify interrupt mask register of 8259
  switch (m_nIRQ)
  {
  case 3:
    outportb(0x21, inportb(0x21) & 0xf7);
    break;
  case 4:
    outportb(0x21, inportb(0x21) & 0xef);
    break;
  }
 
  // Switch to enable baudrate write  
  outportb(m_Regs8250.LCR, 0x80); 

  // Set baud to 9600
  outportb(m_Regs8250.DLH, 0x00);
  outportb(m_Regs8250.DLL, 0x0c);
  
  // Set parity, length and stop bits
  outportb(m_Regs8250.LCR, 
        0x00 | // Disable baudrate write 
        0x00 | // Parity: none
        0x03 | // Length: 8 bits
        0x00   // Stop bits: 1
  );
    
  // Enable interrupt on data available
  outportb(m_Regs8250.IER, 0x01);
    
  // Raise DTR, RTS and OUT2 (respectively)
  outportb(m_Regs8250.MCR, 0x01 | 0x02 | 0x08);
  
  m_bFlowRecv = 1;
  m_bFlowSend = 1;
}

CComPort::~CComPort()
{
  switch (m_nIRQ)
  {
  case 3:
    outportb(0x21, inportb(0x21) | 0x08);
    break;
  case 4:
    outportb(0x21, inportb(0x21) | 0x10);
    break;
  }

  // Disable port and interrupts
  outportb(m_Regs8250.LCR, 0x7f);
  outportb(m_Regs8250.IER, 0x00);

  _go32_dpmi_set_protected_mode_interrupt_vector(m_nInterrupt, &m_OldInfo);
  _go32_dpmi_free_iret_wrapper(&m_NewInfo);
}

int CComPort::ReadData(char *c)
{
  if (m_BufferHead != m_BufferTail) 
  {
    *c = m_Buffer[m_BufferTail];
    m_BufferTail = (m_BufferTail + 1) % COM_PORT_BUFFER_SIZE;
#ifdef COM_PORT_XON_XOFF_ENABLED
    if (!m_bFlowRecv && (BufferLength() < COM_PORT_BUFFER_FLOW_ON))
    {
      WriteData(&XON);
      m_bFlowRecv = 1;
    }
#endif
    return 1;
  }
  else
  {
    return 0;
  }
}

int CComPort::WriteData(const char *c)
{
#ifdef COM_PORT_XON_XOFF_ENABLED
  while (!m_bFlowSend);
#endif
  do
  {
    delay(1);
  }
  while (inportb(m_Regs8250.LSR) & 0x20 != 20);

  outportb(m_Regs8250.THR, *c);
  return 1;
}
