#include <stdio.h>
#include <iostream.h>
#include <conio.h>
#include <sys/stat.h>
#include <time.h>
#include "packet.h"
#include "comport.h"
#include "do.h"

extern CComPort * pComPort;
extern int bDebug;

int KermitGet(int bShowStatus = 1)
{
  LogMessage(LOG_EVENT, "Receiving file");
  ResetEscape();
  
  int done = 0;
  int state = 0;
  int PacketCount = 0;
  
  FILE * OutputFile = NULL; 

  pComPort->Flush();
  
  while (!done)
  {
    Packet CurrentPacket;
    switch (state)
    {
    case 0: // Wait for send-init
      if (bDebug)
        cout << "Waiting for send-init..." << endl;
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'S')
        {
          SendPacket('Y', "~* @-#N1", 0);
          PacketCount = 1;
          state = 1;
        } else if (CurrentPacket.Type == 'E')
        {
          int Size = strlen(CurrentPacket.Data);
          Unescape(CurrentPacket.Data, &Size);
          CRLFToCR(CurrentPacket.Data, &Size);
          cout << "Error: " << CurrentPacket.Data << endl;
          done = 1;
          LogMessage(LOG_ERROR, CurrentPacket.Data);
        }
        else
          ACKPacket(CurrentPacket.Sequence);
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
        else
          NAKPacket(CurrentPacket.Sequence);
      }
      break;
    case 1: // Wait for file-header
      if (bDebug)
        cout << "Waiting for file-header..." << endl;
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'F')
        {
          if (bDebug)
            cout << "File name is " << CurrentPacket.Data << endl;
          cout << "Receiving: " << CurrentPacket.Data << endl;
          LogMessage(LOG_MESSAGE, CurrentPacket.Data);
          OutputFile = fopen(CurrentPacket.Data, "wb");
          PacketCount = 2;
          state = 2;
        } else if (CurrentPacket.Type == 'X')
        {
          strcpy(CurrentPacket.Data, "OUTPUT");
          if (bDebug)
            cout << "File name is " << CurrentPacket.Data << endl;
          LogMessage(LOG_MESSAGE, CurrentPacket.Data);
          OutputFile = fopen(CurrentPacket.Data, "wb");
          PacketCount = 2;
          state = 2;
        } else if (CurrentPacket.Type == 'B')
        {
          done = 1;
        }
        ACKPacket(CurrentPacket.Sequence);
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
        else
          NAKPacket(CurrentPacket.Sequence);
      }
      break;
    case 2: // Wait for file-data
      if (bDebug)
        cout << "Waiting for file-data..." << endl;
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'D')
        {
          gotoxy(1, wherey());
          clreol();
          if (bShowStatus)
            cout << "Packet: " << CurrentPacket.Sequence << " (" << PacketCount << ")" << flush;
          if (CurrentPacket.Sequence == PacketCount)
          {
            int Size = strlen(CurrentPacket.Data);
            Unescape(CurrentPacket.Data, &Size);
            CRLFToCR(CurrentPacket.Data, &Size);
            PacketCount = (PacketCount + 1) % 64;
            fwrite(CurrentPacket.Data, 1, Size, OutputFile);
          }
          else
          {
            LogMessage(LOG_ERROR, "Out of sequence packet");
          }
        }
        else 
        {
          gotoxy(1, wherey());
          clreol();
          if (CurrentPacket.Type == 'B' || CurrentPacket.Type == 'E')
          {
            done = 1;
            if (bShowStatus)
              cout << "Aborted";
          }
          else if (CurrentPacket.Type == 'Z')
          {
            state = 1;
            if (bShowStatus)
              cout << "Finished";
          }
          LogMessage(LOG_EVENT, "Successful");
          fflush(OutputFile);
          struct stat FileInfo;
          fstat(fileno(OutputFile), &FileInfo);
          if (bShowStatus)
            cout << ", received " << FileInfo.st_size << " byte(s)" << endl;
          if (bDebug)
            cout << "Closing file" << endl;
          fclose(OutputFile);
        }
        ACKPacket(CurrentPacket.Sequence);
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
        else
          NAKPacket(CurrentPacket.Sequence);
      }
      break;
    }
  }
  
  return 1;
}

int KermitPut(const char * Filename)
{
  LogMessage(LOG_EVENT, "Sending file");
  ResetEscape();
  
  int done = 0;
  int state = 0;
  int PacketCount = 0, FileOffset = 0;
  char RemoteFileName[64];  

  FILE * OutputFile = fopen(Filename, "rb"); 
  if (!OutputFile)  
  {
    cout << "Error: File could not be opened" << endl;
    return 0;
  }

  struct stat FileInfo;
  fstat(fileno(OutputFile), &FileInfo);

  int Pos;
  for (Pos = strlen(Filename) - 1; Pos >= 0; Pos--)
  {
    if (Filename[Pos] == '\\' || Filename[Pos] == '/' || Filename[Pos] == ':')
      break;
  }
  Filename += Pos + 1;

  pComPort->Flush();
  unsigned int StartTime = rawclock();

  while (!done)
  {
    Packet CurrentPacket;
    switch (state)
    {
    case 0: // Send send-init
      if (bDebug)
        cout << "Waiting for ack send-init..." << endl;
      SendPacket('S', "~* @-#N1", 0);
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'Y')
        {
          state = 1;
        }
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
      }
      break;
    case 1: // Send file-header
      if (bDebug)
        cout << "Waiting for ack file-header..." << endl;
      SendPacket('F', Filename, 1);
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'Y' && CurrentPacket.Sequence == 1)
        {
          strcpy(RemoteFileName, CurrentPacket.Data);
          state = 2;
          PacketCount = 2;
        }
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
      }
      break;
    case 2: // Sending file-data
      if (bDebug)
        cout << "Waiting for ack file-data..." << endl;
      
      {
        fseek(OutputFile, FileOffset, SEEK_SET);
        const int BLOCK_SIZE = 45;
        
        int Size = BLOCK_SIZE, ReadSize; 
        Size = ReadSize = fread(CurrentPacket.Data, 1, Size, OutputFile);
        CurrentPacket.Data[Size] = '\0';
        Escape(CurrentPacket.Data, &Size);
        SendPacket('D', CurrentPacket.Data, PacketCount);
        
        if (GetPacket(&CurrentPacket))
        {
          if (CurrentPacket.Type == 'Y' && CurrentPacket.Sequence == PacketCount)
          {
            if (!bDebug)
            {
              gotoxy(1, wherey());
              clreol();
              cout << "Packet: " << PacketCount << " (" << int(FileOffset * 100 / FileInfo.st_size) << "%";
              if (FileOffset > 1000)
              {
                int CurrentTime = rawclock();
                double FinishRaw = (double)(CurrentTime - StartTime) * FileInfo.st_size / FileOffset - (CurrentTime - StartTime);
                int FinishTime = int(FinishRaw / 18.2);
                cout << ", " << FinishTime << " secs remain";
              }
              
              cout << ")" << flush;
            }
            PacketCount = (PacketCount + 1) % 64;
            FileOffset += ReadSize;
            if (ReadSize < BLOCK_SIZE)
              state = 3;
          }
        }
        else
        {
          if (HitEscape())
          {
            SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
            return 0;
          }
        }
      }
      break;
    case 3: // Sending break
      if (bDebug)
        cout << "Waiting for ack EOF..." << endl;
      if (!bDebug)
      {
        gotoxy(1, wherey());
        clreol();
        cout << "Finishing" << endl;
      }
      SendPacket('Z', "", PacketCount);
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'Y' && CurrentPacket.Sequence == PacketCount)
        {
          PacketCount = (PacketCount + 1) % 64;
          state = 4;
        }
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
      }
      break;
    case 4: // Sending break
      if (bDebug)
        cout << "Waiting for ack break..." << endl;
      SendPacket('B', "", PacketCount);
      if (GetPacket(&CurrentPacket))
      {
        if (CurrentPacket.Type == 'Y' && CurrentPacket.Sequence == PacketCount)
          done = 1;
      }
      else
      {
        if (HitEscape())
        {
          SendPacket('E', "Interrupted#M#J", CurrentPacket.Sequence);
          return 0;
        }
      }
      break;
    }
  }
  
  if (!bDebug)
  {
    gotoxy(1, wherey());
    clreol();
  }
  else
    cout << "Closing file" << endl;

  fflush(OutputFile);
  cout << "Finished, sent " << FileInfo.st_size << " byte(s) to " << RemoteFileName << endl;
  fclose(OutputFile);
  return 1;
}
