#include <fstream.h>
#include <iostream.h>
#include <conio.h>
#include "comport.h"
#include "packet.h"

extern CComPort * pComPort;
extern int bDebug;

void ACKPacket(int PacketNum)
{
  SendPacket('Y', "", PacketNum);
}

void NAKPacket(int PacketNum)
{
  SendPacket('N', "", PacketNum);
}

void Escape(char * Data, int * Size)
{
  int i, j;  
  for (i = 0; i < *Size; i++)
  {
    if (Data[i] == '#' || (((unsigned short int)Data[i] & 127) < 32) ||
      (((unsigned short int)Data[i] & 127) == 127))
    {
      for (j = *Size; j > i; j--)
      {
        Data[j] = Data[j - 1];
      }
      (*Size)++;
      Data[i] = '#';
      i++;
      if (Data[i] != '#')
      {
        Data[i] = ctl(Data[i]);
      }
    }
  }
  Data[*Size] = '\0';
}

void Unescape(char * Data, int * Size)
{
  int i, j;  
  for (i = 0; i < *Size - 1; i++)
  {
    if (Data[i] == '#')
    {
      for (j = i; j < *Size - 1; j++)
      {
        Data[j] = Data[j + 1];
      }
      (*Size)--;
      if (Data[i] != '#')
      {
        Data[i] = ctl(Data[i]);
      }
    }
  }
  Data[*Size] = '\0';
}

void CRLFToCR(char * Data, int * Size)
{
  int i, j;  
  for (i = 0; i < *Size - 1; i++)
  {
    if (Data[i] == 13 && Data[i + 1] == 10)
    {
      for (j = i; j < *Size - 1; j++)
      {
        Data[j] = Data[j + 1];
      }
      (*Size)--;
    }
  }
  Data[*Size] = '\0';
}

int bHitEscape = 0;

void ResetEscape()
{
  bHitEscape = 0;
}

int HitEscape()
{
  return bHitEscape;
}

int GetChar(char * c)
{
  if (bHitEscape)
    return 0;

  while (1)
  {
    if (pComPort->ReadData(c))
    {
      return 1;
    }
    if (kbhit())
    {
      char c = getch();
      if (c == 27)
      {
        if (bDebug)
          cout << "ESC hit!" << endl;
        bHitEscape = 1;
        return 0;
      }
    }
  }
}

int CalcPacketChecksum1(Packet * pPacket)
{
  int CheckSum = 0;
  CheckSum += tochar(pPacket->Length);
  CheckSum += tochar(pPacket->Sequence);
  CheckSum += pPacket->Type;
  for (int i = 0; i < pPacket->Length - 3; i++)
  {
    CheckSum += pPacket->Data[i];
  }
  CheckSum = (CheckSum + ((CheckSum & 192) / 64)) & 63;

  return CheckSum;
}

void PrintPacket(Packet * pPacket)
{
  if (!bDebug)
    return;
  pPacket->CalcCheck = CalcPacketChecksum1(pPacket);
  cout << "Packet #" << pPacket->Sequence 
       << ": Type = " << pPacket->Type 
       << ", Length = " << pPacket->Length
       << ", Check = " << pPacket->Check 
       << " (" << pPacket->CalcCheck << ")" << endl
       << "Data: ";
  cout << pPacket->Data << endl;
}

void SendPacket(Packet * pPacket)
{
  if (bDebug)
    cout << "Sending packet: " << endl;
  PrintPacket(pPacket);
  char c;
  c = 1;
  pComPort->WriteData(&c);
  c = tochar(pPacket->Length);
  pComPort->WriteData(&c);
  c = tochar(pPacket->Sequence);
  pComPort->WriteData(&c);
  c = pPacket->Type;
  pComPort->WriteData(&c);
  for (int i = 0; i < pPacket->Length - 3; i++)
  {
    c = pPacket->Data[i];
    pComPort->WriteData(&c);
  }
  c = tochar(pPacket->Check);
  pComPort->WriteData(&c);
}

void SendPacket(char Type, const char * Data, int SequenceNum)
{
  Packet Pkt;
  Pkt.Length = strlen(Data) + 3;
  Pkt.Sequence = SequenceNum;
  Pkt.Type = Type;
  strcpy(Pkt.Data, Data);
  Pkt.Check = CalcPacketChecksum1(&Pkt);
  SendPacket(&Pkt);
}

int GetPacket(Packet * pPacket)
{
  int done = 0; 
  while (!done)
  {
    char c;
    if (GetChar(&c))
    {
      if (c == 1) 
        done = 1;
    }
    else
      return 0;
  }
  
  char c;   

  if (GetChar(&c))
    pPacket->Length = unchar(c);    
  else
    return 0;

  if (GetChar(&c))
    pPacket->Sequence = unchar(c) & 63;
  else
    return 0;
  
  if (GetChar(&c))
    pPacket->Type = c;    
  else
    return 0;

  int i;

  for (i = 0; i < pPacket->Length - 3; i++)
  {
    if (GetChar(&c))
      pPacket->Data[i] = c;
    else
      return 0;
  }

  pPacket->Data[i] = '\0';

  if (GetChar(&c))
    pPacket->Check = unchar(c);    
  else
    return 0;

  PrintPacket(pPacket);

  pPacket->CalcCheck = CalcPacketChecksum1(pPacket);

  if (pPacket->Check != pPacket->CalcCheck)
  {
    if (bDebug)
      cout << "Checksums don't match: sending NAK" << endl;
    SendPacket('N', "", pPacket->Sequence);
    return 0;
  }

  return 1;
}
